/* SPDX-License-Identifier: GPL-2.0+ OR Apache-2.0 */
/*
 * Copyright (C) 2020 Gao Xiang <hsiangkao@aol.com>
 */
#ifndef __EROFS_TRACE_H
#define __EROFS_TRACE_H

#ifdef __cplusplus
extern "C"
{
#endif

#define trace_erofs_map_blocks_flatmode_enter(inode, map, flags) ((void)0)
#define trace_erofs_map_blocks_flatmode_exit(inode, map, flags, ret) ((void)0)

#ifdef __cplusplus
}
#endif

#endif
